// -*-c++-*-

/*!
  \file obsolete/intention_kick2007.cpp
  \brief queued kick behavior.
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "intention_pass.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/obsolete/body_kick_two_step.h>
#include <rcsc/action/obsolete/body_kick_multi_step.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/say_message_builder.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

/*-------------------------------------------------------------------*/
/*!

 */
IntentionPass::IntentionPass( const rcsc::Vector2D & target_point,
                              const double & first_speed,
                              const int kick_step,
                              const bool enforce_kick,
                              const int & target_unum,
                              const rcsc::GameTime & start_time )
    : M_target_point( target_point )
    , M_first_speed( std::min( first_speed, rcsc::ServerParam::i().ballSpeedMax() ) )
    , M_kick_step( kick_step )
    , M_enforce_kick( enforce_kick )
    , M_target_unum( target_unum )
    , M_last_execute_time( start_time )
    , M_first_kick_step( kick_step )
{

}

/*-------------------------------------------------------------------*/
/*!

 */
bool
IntentionPass::finished( const rcsc::PlayerAgent * agent )
{
    if ( M_kick_step <= 0 )
    {
        rcsc::dlog.addText( rcsc::Logger::KICK,
                            __FILE__": kick intention finished" );
        return true;
    }

    if ( ! agent->world().self().isKickable() )
    {
        return true;
    }

    if ( M_last_execute_time.cycle() + 1 != agent->world().time().cycle() )
    {
        return true;
    }

    if ( agent->world().existKickableOpponent() )
    {
        return true;
    }

    // check near opponents
    const double kickable_plus = rcsc::ServerParam::i().defaultKickableArea() + 0.1;

    const rcsc::PlayerPtrCont::const_iterator o_end = agent->world().opponentsFromBall().end();
    for ( rcsc::PlayerPtrCont::const_iterator it = agent->world().opponentsFromBall().begin();
          it != o_end;
          ++it )
    {
        if ( (*it)->distFromBall() > 5.0 )
        {
            break;
        }

        // check opponent kickable current ball pos
        if ( (*it)->distFromBall() < kickable_plus )
        {
            return true;
        }
        // check opponent goalie
        else if ( (*it)->goalie()
                  && (*it)->pos().x > rcsc::ServerParam::i().theirPenaltyAreaLineX() + 1.5
                  && (*it)->pos().absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth() - 1.5
                  && (*it)->distFromBall() < rcsc::ServerParam::i().catchAreaLength() )
        {
            return true;
        }
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
IntentionPass::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::KICK,
                  __FILE__": execute. target=(%.1f, %.1f) speed=%.3f",
                  M_target_point.x, M_target_point.y, M_first_speed );

    if ( M_kick_step <= 0 )
    {
        rcsc::dlog.addText( rcsc::Logger::KICK,
                      __FILE__": error. empty kick intention" );
        return false;
    }

    
    //say
    if ( agent->config().useCommunication()
         && M_kick_step == M_first_kick_step
         && M_target_unum != -1 )
    {
         rcsc::dlog.addText( rcsc::Logger::PASS,
                             "%s:%d: IntentionPass::finished() set pass communication."
                             ,__FILE__, __LINE__ );
         rcsc::Vector2D target_buf = M_target_point - agent->world().self().pos();
         target_buf.setLength( 1.0 );

         agent->addSayMessage( new rcsc::PassMessage( M_target_unum,
                                                      M_target_point + target_buf,
                                                      agent->effector().queuedNextBallPos(),
                                                      agent->effector().queuedNextBallVel() ) );
    }
        

    if ( M_enforce_kick )
    {
        if ( M_kick_step == 1 )
        {
            // if enforce mode. not consier opponent.
            rcsc::dlog.addText( rcsc::Logger::KICK,
                          __FILE__": execute. enforce one kick" );
            rcsc::Body_KickOneStep( M_target_point, M_first_speed ).execute( agent );
        }
        else if ( M_kick_step == 2 )
        {
            rcsc::dlog.addText( rcsc::Logger::KICK,
                          __FILE__": execute. enforce two kick" );
            rcsc::Body_KickTwoStep( M_target_point,
                              M_first_speed,
                              false
                              ).execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::KICK,
                          __FILE__": execute. enforce, but multi kick" );
            rcsc::Body_KickMultiStep( M_target_point,
                                M_first_speed,
                                false // not enforce
                                ).execute( agent );
        }
    }
    else
    {
        rcsc::Vector2D next_vel;
        const rcsc::Vector2D target_rpos = M_target_point - agent->world().self().pos();

        // SEE: obsolete/body_kick_multi_step::execute()

        // try one step
        if ( rcsc::Body_KickTwoStep::simulate_one_kick
             ( &next_vel,
               NULL,
               NULL,
               target_rpos,
               M_first_speed,
               rcsc::Vector2D(0.0, 0.0), // my current pos
               agent->world().self().vel(),
               agent->world().self().body(),
               agent->world().ball().rpos(),
               agent->world().ball().vel(),
               agent,
               false ) // not enforce
             )
        {
            rcsc::Vector2D required_acc = next_vel - agent->world().ball().vel();
            double kick_pow = required_acc.r() / agent->world().self().kickRate();
            rcsc::AngleDeg kick_dir = required_acc.th() - agent->world().self().body();

            rcsc::dlog.addText( rcsc::Logger::KICK,
                          __FILE__": execute. can reach one step" );
            agent->doKick( kick_pow, kick_dir );
            this->clear(); // finish
        }
        // try two step
        else if ( M_kick_step >=2
                  && rcsc::Body_KickTwoStep::simulate_two_kick
                  ( NULL,
                    &next_vel,
                    target_rpos,
                    M_first_speed,
                    rcsc::Vector2D(0.0, 0.0), // my current pos
                    agent->world().self().vel(),
                    agent->world().self().body(),
                    agent->world().ball().rpos(),
                    agent->world().ball().vel(),
                    agent,
                    false ) // not enforced
                  )
        {
            rcsc::Vector2D required_acc = next_vel - agent->world().ball().vel();
            double kick_pow = required_acc.r() / agent->world().self().kickRate();
            rcsc::AngleDeg kick_dir = required_acc.th() - agent->world().self().body();

            rcsc::dlog.addText( rcsc::Logger::KICK,
                          __FILE__": execute. can reach two step" );
            agent->doKick( kick_pow, kick_dir );
        }
        else if ( M_kick_step >= 3
                  && rcsc::Body_KickMultiStep::simulate_three_kick
                  ( NULL,
                    &next_vel,
                    target_rpos,
                    M_first_speed,
                    rcsc::Vector2D( 0.0, 0.0 ), // my current pos
                    agent->world().self().vel(),
                    agent->world().self().body(),
                    agent->world().ball().rpos(),
                    agent->world().ball().vel(),
                    agent,
                    false ) // not enforced
                  )
        {
            rcsc::Vector2D required_acc = next_vel - agent->world().ball().vel();
            double kick_pow = required_acc.r() / agent->world().self().kickRate();
            rcsc::AngleDeg kick_dir = required_acc.th() - agent->world().self().body();

            rcsc::dlog.addText( rcsc::Logger::KICK,
                          __FILE__": execute. can reach three step" );
            agent->doKick( kick_pow, kick_dir );
        }
        else if ( M_kick_step >= 3 )
        {
            rcsc::Body_HoldBall().execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::KICK,
                          __FILE__": execute. never reach. failed" );
            this->clear();
            return false;
        }
    }

    --M_kick_step;

    // update
    M_last_execute_time = agent->world().time();

    const rcsc::AngleDeg target_angle
        = ( M_target_point
            - ( agent->world().self().pos() + agent->world().self().vel() ) ).th();

    if ( agent->world().dirCount( target_angle ) == 0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToPoint( M_target_point ) );
    }

    return true;
}

