#include "tile_coding.h"
#include <iostream>
#include <cmath>
#include <sstream>

using namespace std;


TileCoding::TileCoding(const int & input_dim,
                       const int & tiling_num,
                       const std::vector<double> & tile_widths)
:M_input_dim((unsigned)input_dim)
,M_tiling_num(tiling_num)
,M_tile_widths(tile_widths)
{
    //error check
    if( M_input_dim != tile_widths.size() ){
        std::cerr << "**ERROR** tile_widths size is incorrect"
                  << "(TileCoding::TileCoding)" << std::endl;
    }
}

std::vector< std::vector<int> >
TileCoding::getTiles( const std::vector<double> & input ) const
{
    std::vector< std::vector<int> > tiles;
    
    if( input.size() != M_input_dim ){
        std::cerr << "**ERROR** input size is incorrect(TileCoding::getTiles)"
                  << std::endl;
        return tiles;
    }

    std::vector<int> qstate,base;
    std::vector<int> coordinates;

    qstate.reserve(M_input_dim);
    base.resize(M_input_dim);
    coordinates.resize(M_input_dim);

    for(unsigned int i=0; i<M_input_dim; ++i){
        qstate.push_back((int)std::floor(input[i]
                                         / M_tile_widths[i] * M_tiling_num));
//        base.push_back(0);
    }

    for(int i=0; i<M_tiling_num; ++i) {
		for(unsigned int j=0; j<M_input_dim; ++j){
            
			if (qstate[j] >= base[j])
				coordinates[j]
                    = qstate[j] - ((qstate[j]-base[j]) % M_tiling_num);
			else
				coordinates[j]
                    = qstate[j]+1 + ((base[j]-qstate[j]-1) % M_tiling_num)
                    - M_tiling_num;
            
			base[j] += 1 + (2*j);
//            base[j] += 1;
            
		}

        tiles.push_back(coordinates);
	}

    return tiles;
}

bool
TileCoding::update( const std::vector<double> & input,
                    const double & diff )
{
    if( input.size() != M_input_dim ){
        std::cerr << "**ERROR** input size is incorrect(TileCoding::update)"
                  << std::endl;
        return false;
    }

    std::vector< std::vector<int> > tiles = getTiles(input);

    for( unsigned int i=0; i<tiles.size(); ++i){
        updateWeight( tiles[i], diff/(double)M_tiling_num );
    }

    return true;
}

double
TileCoding::getValue( const std::vector<double> & input ) const
{
    if( input.size() != M_input_dim ){
        std::cerr << "**ERROR** input size is incorrect(TileCoding::getValue)"
                  << std::endl;
        return 0.0;
    }
    
    std::vector< std::vector<int> > tiles = getTiles(input);

    double sum = 0;

    for( unsigned int i=0; i<tiles.size(); ++i){
        sum += getWeight( tiles[i] );
    }

    return sum;
}

bool
TileCoding::read( const std::string & file_name )
{
    FILE *fp;
    if ((fp = fopen(file_name.c_str(), "r")) == NULL) {
        return false;
    }
        
     char buffer[256];
     char word[128];

     double value=0;
     while (fgets(buffer, 256, fp) != NULL) {
         std::vector<int> key;

         int idx = 0, w_idx = 0;
         while( buffer[idx] != '|' ){
             if( buffer[idx] != ' ' ){
                 word[w_idx] = buffer[idx];
                 ++w_idx;
             } else {
                 word[w_idx] = '\0';
                 w_idx = 0;
                 key.push_back(atoi(word));
             }
             ++idx;
         }
         w_idx = 0;
         idx += 2; // |μμʸɤ߼
         while( buffer[idx] != '\n' ){
             word[w_idx] = buffer[idx];
             ++w_idx;
             ++idx;
         }
         word[w_idx] = '\0';
         
         sscanf( word, "%lf", &value );

         M_weight.insert( pair<std::vector<int>,double>(key,value) );
 	}
    
     printf("finished reading value.\n");
        
 	fclose(fp);

    return true;
}

bool
TileCoding::write( const std::string & file_name ) const
{
     FILE *fp;
     if ((fp = fopen(file_name.c_str(), "w")) == NULL) {
         printf("**WARNING** cant write value\n");
         return false;
     }
        
     std::ostringstream buffer;

     std::map< std::vector<int>, double >::const_iterator it;
     for(it=M_weight.begin(); it!=M_weight.end(); ++it)
     {
         if( std::abs(it->second) > 10e-10 ){
             for(unsigned int i=0; i<(it->first).size(); ++i){
                 buffer << it->first[i] << " ";
             }
             buffer << "| " ;//<< std::endl;
             buffer << it->second << std::endl;
         }
         
         if( buffer.str().size() > 32768 ){
             fputs(buffer.str().c_str(),fp);
             buffer.str("");
             buffer.clear();
         }
     }

     if( buffer.str().size() > 0 ){
         fputs(buffer.str().c_str(),fp);
     }
     
     fclose(fp);
     
     return true;
}
